%EIMM_DEMO2 Bearings Only Tracking of a Manouvering Target demonstration
% 
% Simple demonstration for non-linear IMM using the following models:
%  1. Standard Wiener process velocity model
%  2. Coordinated turn model
%
% The measurement model is non-linear bearings only model.
% See BOT-demo or the documentation for more details.
% 
% Copyright (C) 2007-2008 Jouni Hartikainen
%
% This software is distributed under the GNU General Public 
% Licence (version 2 or later); please refer to the file 
% Licence.txt, included with the software, for details.


%% 4 Fitting approaches are added by Tiancheng Li (t.c.li@usal.es)

% Continuous time trajectory Function of Time (FoT) fitting based 
%              Smoothing (Delayed fitting):  - line 399
%              Filtering (Online fitting):   - line 402
%              Forecasting:                  - line 405
%          and Smoothed fitting  :           - line 409

% For the technical details, please refer to the paper:
% T. Li, H. Chen, S. Sun and J.M. Corchado, Joint Smoothing, Tracking, and Forecasting Based
% on Continuous-Time Target Trajectory Fitting, arXiv:1708.02196 [stat.AP]
% or the website: https://sites.google.com/site/tianchengli85/o2

%%
clear
print_figures = 1;
save_figures = 1;

% Dimensionality of the state space
dims = 5;

% The number of models in use
nmodels = 2;

% Step size
dt = 0.1;

% Space for function handles and parameters
a_func = {};
ia_func = {};
a_param = {};
h_func = {};
h_param = {};
dh_dx_func = {};

a_func{1} = [];
a_func{2} = @f_turn;
ia_func{1} = [];
ia_func{2} = @f_turn_inv;
a_param{1} = [];
a_param{2} = {dt};


% Space for model parameters
ind = cell(1,nmodels);
F   = cell(1,nmodels);
L   = cell(1,nmodels);
Qc  = cell(1,nmodels);
A   = cell(1,nmodels);
Q   = cell(1,nmodels);
H   = cell(1,nmodels);
R   = cell(1,nmodels);

% Index vector of model 1
ind{1} = [1 2 3 4]';

% Transition matrix for the continous-time velocity model.
F{1} = [0 0 1 0;
        0 0 0 1;
        0 0 0 0;
        0 0 0 0];

% Noise effect matrix for the continous-time system.
L{1} = [0 0;
        0 0;
        1 0;
        0 1];

% Process noise variance
q1 = 0.01;
Qc{1} = diag([q1 q1]);

% Discretization of the continous-time system.
[A{1},Q{1}] = lti_disc(F{1},L{1},Qc{1},dt);

% Process noise variance for model 1 using EKF
EKF_q1 = .05;
EKF_Qc1 = diag([EKF_q1 EKF_q1]);

% Discretization of the continous-time system.
[EKF_A1,EKF_Q1] = lti_disc(F{1},L{1},EKF_Qc1,dt);


%%% Specification of the turning model

% System components. 5th parameter is the turning rate 
ind{2} = [1 2 3 4 5]';

% Derivative of the dynamic function 
A{2} = @f_turn_dx;
% Process noise for the turning rate
Qc{2} = 0.15;

% Noise effect matrix
L{2} = [0 0 0 0 1]';

% Process noise covariance
Q{2} = L{2}*Qc{2}*L{2}'*dt;

hdims = 2;

%mu_ip = [0.90 0.05 0.05];
mu_ip = [0.95 0.05];
mu_0j = mu_ip;
%p_ij = [0.65 0.35;
%        0.10 0.90];

p_ij = [0.90 0.10;
        0.10 0.90];

% Number of data points
n = 200;

% cpu time record
EKFtime = 0;
UKFtime = 0;
EIMMtime = 0;
UIMMtime = 0;

EKStime = 0;
UKStime = 0;
EIMMStime = 0;
UIMMStime = 0;

UimmPretime = 0;

delayFtime = 0;
filterFtime = 0;
smoothFtime = 0;
predictFtime = 0;

MCruns = 200; T = 1:n; predictStep = 5;
SensorNoise_Factor = 2; %  using factor 1, just normal processing with everything perfect assumed by the estimator
% using the other to change the real sensor noise without informing any estimators

for rep = 1: MCruns
    % Space for real states and modes
    X_r = zeros(dims,n);
    mstate = zeros(1,n);

    %%%%%%% Creation of trajectory %%%%%%%

    % Start with constant velocity 1 toward right
    mstate(1:40) = 1;
    X_r(:,1) = [0 0 1 0 0]';

    % At 4s make a turn left with rate 1 
    mstate(41:90) = 2;
    X_r(5,40) = 1;

    % At 9s move straight for 2 seconds
    mstate(91:110) = 1;

    % At 11s commence another turn right with rate -1
    mstate(111:160) = 2;
    X_r(5,110) = -1;

    % At 16s move straight for 4 seconds
    mstate(161:200) = 1;

    % Generate object state values
    for i = 2:n
       st = mstate(i);
       if isstr(a_func{st}) | strcmp(class(a_func{st}),'function_handle')
           X_r(ind{st},i) = feval(a_func{st},X_r(ind{st},i-1),a_param{st});
       else 
           X_r(ind{st},i) = A{st}*X_r(ind{st},i-1);
       end
    end

    % Positions of sensors
    % $$$ S1 = [-0.5; 3];
    % $$$ S2 = [-0.5;-3];
    % $$$ S3 = [   7;-3];
    % $$$ S4 = [   7; 3];

    S1 = [-0.5; 3.5];
    S2 = [-0.5;-3.5];
    S3 = [   7;-3.5];
    S4 = [   7; 3.5];

    s = [S1 S2 S3 S4];

    % Handles to measurement models
    H{1} = @bot_dh_dx;
    H{2} = @bot_dh_dx;

    h_func{1} = @bot_h;
    h_func{2} = @bot_h;

    h_param{1} = s;
    h_param{2} = s;

    % Standard deviation
    sd = 0.1*ones(1,size(s,2));
    R = {};
    R{1} = diag(sd.^2);
    R{2} = R{1};
    % Generate measurements
    Y = bot_h(X_r,s);
    % Add noise
    for i = 1:size(Y,2)
        st = mstate(i);
        for j = 1:size(Y,1)
            Y(j,i) = Y(j,i) + sqrt(R{st}(j,j)) * randn/SensorNoise_Factor; % !! Adapting the sensor noise without telling ESTIMATORS
        end
    end

%     % Print the trajectory
%     if print_figures
%         h = plot(X_r(1,:),X_r(2,:),'-g',...
%                  s(1,:),s(2,:),'k^',... 
%                  X_r(1,1),X_r(2,1),'ro','MarkerSize',12);    
%         legend('Real trajectory',...
%                'Positions of sensors',...
%                'Starting position', 'Location', 'North');
%         set(h,'markersize',5);
%         set(h,'linewidth',0.5);
%         set(gca,'FontSize',8);
%         xlim([-1 7.5]) 
%         ylim([-4 4]) 
%         if save_figures
%             print('-depsc','eimm2_trajectory.eps');
%         end
%         pause
%     end


    m = [0 0 0 -1 0]';
    P = diag([10.1 10.1 1.1 1.1 1]);

    %% Space for the estimates.

    % EKF with model 1
    EKF_MM = zeros(size(A{1},1), size(Y,2));
    EKF_PP = zeros(size(A{1},1), size(A{1},1), size(Y,2));

    % UKF with model 1
    UKF_MM = zeros(size(A{1},1), size(Y,2));
    UKF_PP = zeros(size(A{1},1), size(A{1},1), size(Y,2));

    % EKF based IMM
    EIMM_MM = zeros(size(m,1), size(Y,2));
    EIMM_PP = zeros(size(m,1), size(m,1), size(Y,2));
    EIMM_MM_i = cell(nmodels,n);
    EIMM_PP_i = cell(nmodels,n);
    EIMM_MU = zeros(nmodels,size(Y,2));

    % UKF based IMM
    UIMM_MM = zeros(size(m,1), size(Y,2));
    UIMM_PP = zeros(size(m,1), size(m,1), size(Y,2));
    UIMM_MM_i = cell(nmodels,n);
    UIMM_PP_i = cell(nmodels,n);
    UIMM_MU = zeros(nmodels,size(Y,2));

    %%% Initial estimates %%%

    % EKF with model 1
    EKF_M = [0 0 0 -1]';
    EKF_P = diag([1.1 1.1 0.1 0.1]);

    % UKF with model 1
    UKF_M = [0 0 0 -1]';
    UKF_P = diag([1.1 1.1 0.1 0.1]);

    % EKF based IMM
    x_ip1{1} = [0 0 1 0]';
    x_ip1{2} = [0 0 1 0 0]';
    mu_ip1 = mu_ip;

    P_ip1{1} = diag([0.1 0.1 0.1 0.1]);
    P_ip1{2} = diag([0.1 0.1 0.1 0.1 1]);

    % UKF based IMM
    x_ip2{1} = [0 0 1 0]';
    x_ip2{2} = [0 0 1 0 0]';
    mu_ip2 = mu_ip;

    P_ip2{1} = diag([0.1 0.1 0.1 0.1]);
    P_ip2{2} = diag([0.1 0.1 0.1 0.1 1]);


    % Filtering steps.
    tic
    for i = T
        % EKF with model 1
        [EKF_M,EKF_P] = kf_predict(EKF_M,EKF_P,EKF_A1,EKF_Q1);
        [EKF_M,EKF_P] = ekf_update1(EKF_M,EKF_P,Y(:,i),H{1},R{1},h_func{1},[],h_param{1});

        EKF_MM(:,i)   = EKF_M;
        EKF_PP(:,:,i) = EKF_P;
    end
    EKFtime = EKFtime+ toc;
    
    tic
    for i = T
        % UKF with model 1
        [UKF_M,UKF_P] = kf_predict(UKF_M,UKF_P,EKF_A1,EKF_Q1);
        [UKF_M,UKF_P] = ukf_update1(UKF_M,UKF_P,Y(:,i),h_func{1},R{1},h_param{1});

        UKF_MM(:,i)   = UKF_M;
        UKF_PP(:,:,i) = UKF_P;
    end
    UKFtime = UKFtime+ toc;
    
    tic
    for i = T
        % EKF based IMM
        [x_p1,P_p1,c_j1] = eimm_predict(x_ip1,P_ip1,mu_ip1,p_ij,ind,dims,A,a_func,a_param,Q);
        [x_ip1,P_ip1,mu_ip1,m1,P1] = eimm_update(x_p1,P_p1,c_j1,ind,dims,Y(:,i),H,h_func,R,h_param);
        EIMM_MM(:,i)   = m1;
        EIMM_PP(:,:,i) = P1;
        EIMM_MU(:,i)   = mu_ip1';
        EIMM_MM_i(:,i) = x_ip1';
        EIMM_PP_i(:,i) = P_ip1';
    end
    EIMMtime = EIMMtime + toc;

    UA = cell(3,n);  
    tic
    for i = T
        % UKF based IMM    
        [x_p2,P_p2,c_j2] = uimm_predict(x_ip2,P_ip2,mu_ip2,p_ij,ind,dims,A,a_func,a_param,Q);
        [x_ip2,P_ip2,mu_ip2,m2,P2] = uimm_update(x_p2,P_p2,c_j2,ind,dims,Y(:,i),H,h_func,R,h_param);

        UIMM_MM(:,i)   = m2;
        UIMM_PP(:,:,i) = P2;
        UIMM_MU(:,i)   = mu_ip2';
        UIMM_MM_i(:,i) = x_ip2';
        UIMM_PP_i(:,i) = P_ip2';
        UA{1,i}= x_ip2; UA{2,i}= P_ip2; UA{3,i}= mu_ip2; 
    end
    UIMMtime = UIMMtime + toc;
    
 UimmPre = zeros(size(m,1), size(Y,2)+predictStep);    
    tic
    for i = T
        a = UA{1,i}; b = UA{2,i}; c= UA{3,i}; 
        for tt = 1:predictStep
            [a,b,c,XX,PP] = uimm_predict(a,b,c,p_ij,ind,dims,A,a_func,a_param,Q);
        end
        UimmPre(:,predictStep+i) = XX;
    end
    UimmPretime = UimmPretime+ toc;
%% smooth
    % Smooth with EKF based IMM smoother
    tic
    [SMI_1,SPI_1,SMI_i_1,SPI_i_1,MU_S_1] = eimm_smooth(EIMM_MM,EIMM_PP,EIMM_MM_i,EIMM_PP_i,EIMM_MU,p_ij,mu_0j,ind,dims,A,ia_func,a_param,Q,R,H,h_func,h_param,Y);
    EIMMStime = EIMMStime + toc;
    
    % Smooth with UKF based IMM smoother
    tic
    [SMI_2,SPI_2,SMI_i_2,SPI_i_2,MU_S_2] = uimm_smooth(UIMM_MM,UIMM_PP,UIMM_MM_i,UIMM_PP_i,UIMM_MU,p_ij,mu_0j,ind,dims,A,ia_func,a_param,Q,R,H,h_func,h_param,Y);
    UIMMStime = UIMMStime+toc;
    
    % Smooth the EKF estimates with RTS smoother using model 1
    tic
    [SM1,SP1] = rts_smooth(EKF_MM,EKF_PP,EKF_A1,EKF_Q1);
    EKStime = EKStime+toc;

    % Smooth the UKF estimates with RTS smoother using model 1
    tic
    [SM2,SP2] = rts_smooth(UKF_MM,UKF_PP,EKF_A1,EKF_Q1);
    UKStime = UKStime+toc;

    warning off
%% STF Fitting 
% parameter
    Parameter.initial = m; Parameter.Band = 10; Parameter.predictStep = predictStep;
% 2nd order fitting function; larger Parameter.Band will be required
% Fun2 = @(x,t) [atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,1),x(1)+x(2)*t+x(3)*t.^2-s(1,1));
%     atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,2),x(1)+x(2)*t+x(3)*t.^2-s(1,2));
%     atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,3),x(1)+x(2)*t+x(3)*t.^2-s(1,3)); 
%     atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,4),x(1)+x(2)*t+x(3)*t.^2-s(1,4));]; %
% %     each row represents a bearing only sensor - 2-order ; 

% 1st order fitting function
Fun1 = @(x,t) [atan2(x(3)+x(4)*t-s(2,1),x(1)+x(2)*t-s(1,1));
    atan2(x(3)+x(4)*t-s(2,2),x(1)+x(2)*t-s(1,2));
    atan2(x(3)+x(4)*t-s(2,3),x(1)+x(2)*t-s(1,3)); 
    atan2(x(3)+x(4)*t-s(2,4),x(1)+x(2)*t-s(1,4));]; % each row represents a bearing only sensor - 1st order
    tic
    xfit_delay = nlinSTFfitting(Y,Fun1,'smoothing',Parameter); % 
    delayFtime =  toc;
    tic
    xfit_filter = nlinSTFfitting(Y,Fun1,'filtering',Parameter);
    filterFtime = toc;
    tic
    xfit_predict = nlinSTFfitting(Y,Fun1,'forecasting',Parameter);
    predictFtime = toc;
    tic
    Parameter.Band = 7; Parameter.FitOrder = 1; % reduce somehow the fitting iterval
    xfit_smooth = STFfitting(xfit_delay(:,end:-1:1),'smoothing',Parameter); % smoothing in time-reverse way
    xfit_smooth = xfit_smooth(:,end:-1:1);
    smoothFtime = smoothFtime + toc;

% Calculate the MSEs
    MSE_EKF(rep,:) =  (X_r(1,:)-EKF_MM(1,:)).^2+ (X_r(2,:)-EKF_MM(2,:)).^2;
    MSE_EKS(rep,:) = (X_r(1,:)-SM1(1,:)).^2 + (X_r(2,:)-SM1(2,:)).^2;
    MSE_UKF(rep,:) =  (X_r(1,:)-UKF_MM(1,:)).^2+ (X_r(2,:)-UKF_MM(2,:)).^2;
    MSE_UKS(rep,:) = (X_r(1,:)-SM2(1,:)).^2 + (X_r(2,:)-SM2(2,:)).^2;
    MSE_EIMM(rep,:) = (X_r(1,:)-EIMM_MM(1,:)).^2 + (X_r(2,:)-EIMM_MM(2,:)).^2;
    MSE_EIMMS(rep,:) = (X_r(1,:)-SMI_1(1,:)).^2+(X_r(2,:)-SMI_1(2,:)).^2 ;
    MSE_UIMM(rep,:) = (X_r(1,:)-UIMM_MM(1,:)).^2+ (X_r(2,:)-UIMM_MM(2,:)).^2;
    MSE_UIMMS(rep,:) = (X_r(1,:)-SMI_2(1,:)).^2+(X_r(2,:)-SMI_2(2,:)).^2 ;
    MSE_UIMMpre(rep,:) = (X_r(1,predictStep+1:end)-UimmPre(1,predictStep+1:end-predictStep)).^2+ (X_r(2,predictStep+1:end)-UimmPre(2,predictStep+1:end-predictStep)).^2;
    MSE_delayF(rep,:) =(X_r(1,:)-xfit_delay(1,:)).^2 + (X_r(2,:)-xfit_delay(2,:)).^2;
    MSE_filterF(rep,:) = (X_r(1,:)-xfit_filter(1,:)).^2+ (X_r(2,:)-xfit_filter(2,:)).^2;
    MSE_predictF(rep,:) = (X_r(1,Parameter.predictStep+1:end)-xfit_predict(1,Parameter.predictStep+1:end-Parameter.predictStep)).^2+(X_r(2,Parameter.predictStep+1:end)-xfit_predict(2,Parameter.predictStep+1:end-Parameter.predictStep)).^2 ;
    MSE_smoothF(rep,:) = (X_r(1,:)-xfit_smooth(1,:)).^2+ (X_r(2,:)-xfit_smooth(2,:)).^2;   
end

meanRMSE_EKF = sqrt(mean(MSE_EKF));
meanRMSE_EKS = sqrt(mean(MSE_EKS));
meanRMSE_UKF = sqrt(mean(MSE_UKF));
meanRMSE_UKS = sqrt(mean(MSE_UKS));
meanRMSE_EIMM = sqrt(mean(MSE_EIMM));
meanRMSE_EIMMS = sqrt(mean(MSE_EIMMS));
meanRMSE_UIMM = sqrt(mean(MSE_UIMM));
meanRMSE_UIMMS = sqrt(mean(MSE_UIMMS));
meanRMSE_UIMMpre = sqrt(mean(MSE_UIMMpre));

meanRMSE_delayF = sqrt(mean(MSE_delayF));
meanRMSE_filterF = sqrt(mean(MSE_filterF));
meanRMSE_predictF = sqrt(mean(MSE_predictF));
meanRMSE_smoothF = sqrt(mean(MSE_smoothF));

save data_botm_demo_200_Rhalf

fprintf('============== The sensor noise is divided by SensorNoise_Factor= %.2f\n',SensorNoise_Factor);
fprintf('Only when SensorNoise_Factor == 1: all estimators are provided with the correct sensor statistics');


fprintf('Root Mean square errors of position estimates:\n');
fprintf('EKF-RMSE = %.4f\n',mean(meanRMSE_EKF));
fprintf('EKS-RMSE = %.4f\n',mean(meanRMSE_EKS));
fprintf('UKF-RMSE = %.4f\n',mean(meanRMSE_UKF));
fprintf('UKS-RMSE = %.4f\n',mean(meanRMSE_UKS));
fprintf('EIMM-RMSE = %.4f\n',mean(meanRMSE_EIMM));
fprintf('EIMMS-RMSE = %.4f\n',mean(meanRMSE_EIMMS));
fprintf('UIMM-RMSE = %.4f\n',mean(meanRMSE_UIMM));
fprintf('UIMMS-RMSE = %.4f\n',mean(meanRMSE_UIMMS));
fprintf('UIMMpre-RMSE = %.4f\n',mean(meanRMSE_UIMMpre));

fprintf('Online Fitting RMSE = %.4f\n',mean(meanRMSE_filterF));
fprintf('Delayed Fitting RMSE = %.4f\n',mean(meanRMSE_delayF));
fprintf('Smoothed Fitting RMSE = %.4f\n',mean(meanRMSE_smoothF));
fprintf('Fitting-based Forecasting RMSE = %.4f\n',mean(meanRMSE_predictF));


fprintf('Computing time:\n');
fprintf('EKF-time = %.4f\n',EKFtime/MCruns);
fprintf('EKS-time = %.4f\n',(EKFtime+EKStime)/MCruns);
fprintf('UKF-time = %.4f\n',UKFtime/MCruns);
fprintf('UKS-time = %.4f\n',(UKFtime+UKStime)/MCruns);
fprintf('EIMM-time = %.4f\n',EIMMtime/MCruns);
fprintf('EIMMS-time = %.4f\n',(EIMMtime+EIMMStime)/MCruns);
fprintf('UIMM-time = %.4f\n',UIMMtime/MCruns);
fprintf('UIMMS-time = %.4f\n',(UIMMtime+UIMMStime)/MCruns);
fprintf('UIMMpre-time = %.4f\n',(UIMMtime + UimmPretime)/MCruns);

fprintf('Online Fitting time = %.4f\n',filterFtime/MCruns);
fprintf('Delayed Fitting time = %.4f\n',delayFtime/MCruns);
fprintf('Smoothed Fitting time = %.4f\n',(smoothFtime+delayFtime)/MCruns);
fprintf('Fitting-based Forecasting time = %.4f\n',predictFtime/MCruns);


% Plot the final filtering and smoothing results
figure
subplot(2,2,1),h0 = title('Truth'); hold on
h1 = plot(X_r(1,:),X_r(2,:),'k-',X_r(1,1),X_r(2,1),'r^',X_r(1,end),X_r(2,end),'rs',s(1,:),s(2,:),'o','markersize',4,'linewidth',2);
legend(h1,'True Trajectory','Trjajectory Starting Point','Trajectory Ending Point','Bearing-only Sensor'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
xlim([-1 7.5]); ylim([-3.5 3.5]) 

% figure
subplot(2,2,2),h0 = title('Filtering'); hold on
plot(X_r(1,:),X_r(2,:),'k-')
h1 = plot(EKF_MM(1,:),EKF_MM(2,:),'g--',UKF_MM(1,:),UKF_MM(2,:),'r--',EIMM_MM(1,:),EIMM_MM(2,:),'c--',UIMM_MM(1,:),UIMM_MM(2,:),'m--',xfit_filter(1,:),xfit_filter(2,:),'b-','markersize',2,'linewidth',2);
legend(h1,'EKF','UKF','EKF-IMM','UKF-IMM','Online Fitting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
xlim([-1 7.5]); ylim([-3.5 3.5]) 

% figure
subplot(2,2,3),h0 = title('Smoothing');hold on
plot(X_r(1,:),X_r(2,:),'k-'); 
h1 = plot(SM1(1,:),SM1(2,:),'--g',SM2(1,:),SM2(2,:),'--r',SMI_1(1,:),SMI_1(2,:),'--c',SMI_2(1,:),SMI_2(2,:),'--m',xfit_delay(1,:),xfit_delay(2,:),'r-',xfit_smooth(1,:),xfit_smooth(2,:),'b-','markersize',2,'linewidth',2);%
legend(h1,'EKF Smoothing','UKF Smoothing','EKF-IMM Smoothing','UKF-IMM Smoothing','Delayed Fitting','Smoothed Fitting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman'); %
xlim([-1 7.5]); ylim([-3.5 3.5]) 

% figure
subplot(2,2,4),h0 = title('Forecasting'); hold on
plot(X_r(1,:),X_r(2,:),'k-'),%
h1 = plot(UimmPre(1,predictStep+1:end),UimmPre(2,predictStep+1:end),'m-->',xfit_predict(1,Parameter.predictStep+1:end),xfit_predict(2,Parameter.predictStep+1:end),'b->','markersize',2,'linewidth',2);
legend(h1,'UKF-IMM-based 5-step Forecasting','Fitting-based 5-step Forecasting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
xlim([-1 7.5]); ylim([-3.5 3.5]) 

figure

h0=plot(T,meanRMSE_EKF,'g--',T,meanRMSE_UKF,'k--',T,meanRMSE_EIMM,'c--',T,meanRMSE_UIMM,'m--',T,meanRMSE_filterF,'b--',...
    T,meanRMSE_EKS,'g-',T,meanRMSE_UKS,'k-',T,meanRMSE_EIMMS,'c-',T,meanRMSE_UIMMS,'m-',T,meanRMSE_delayF,'r-',T,meanRMSE_smoothF,'b-',...
    T(predictStep+1:end),meanRMSE_UIMMpre,'k:',T(predictStep+1:end),meanRMSE_predictF,'b:'); % 
set(gca,'FontSize',12);set(h0,'LineWidth',2);
h1=xlabel('Sampling steps'); h2=ylabel('RMSE');
h3=legend('EKF','UKF','EKF-IMM','UKF-IMM','Online Fitting','EKF Smoothing','UKF Smoothing','EKF-IMM Smoothing','UKF-IMM Smoothing','Delayed Fitting','Smoothed Fitting',...
   'UKF-IMM-based 5-step Forecasting','Fitting-based 5-step Forecasting', 'location','best');%,'Orientation','horizontal'
set([h1,h2],'FontSize',14,'FontName','Times new Roman');%set(h3,'FontSize',12,'FontName','Times new Roman');
