%EIMM_DEMO2 Bearings Only Tracking of a Manouvering Target demonstration
% 
% Simple demonstration for non-linear IMM using the following models:
%  1. Standard Wiener process velocity model
%  2. Coordinated turn model
%
% The measurement model is non-linear bearings only model.
% See BOT-demo or the documentation for more details.
% 
% Copyright (C) 2007-2008 Jouni Hartikainen
%
% This software is distributed under the GNU General Public 
% Licence (version 2 or later); please refer to the file 
% Licence.txt, included with the software, for details.


%% 4 Fitting approaches are added by Tiancheng Li (t.c.li@usal.es)

% Continuous time trajectory Function of Time (FoT) fitting based 
%              Smoothing (Delayed fitting):  - line 399
%              Filtering (Online fitting):   - line 402
%              Forecasting:                  - line 405
%          and Smoothed fitting  :           - line 409

% For the technical details, please refer to the paper:
% T. Li, H. Chen, S. Sun and J.M. Corchado, Joint Smoothing, Tracking, and Forecasting Based
% on Continuous-Time Target Trajectory Fitting, arXiv:1708.02196 [stat.AP]
% or the website: https://sites.google.com/site/tianchengli85/o2

%%
clear
print_figures = 1;
save_figures = 1;

% Dimensionality of the state space
dims = 4;

% The number of models in use
nmodels = 2;

% Step size
dt = 0.01;

% Space for function handles and parameters
a_func = {};
ia_func = {};
a_param = {};
h_func = {};
h_param = {};
dh_dx_func = {};

a_func{1} = [];
a_func{2} = @f_turn;
ia_func{1} = [];
ia_func{2} = @f_turn_inv;
a_param{1} = [];
a_param{2} = {dt};


% Space for model parameters

% Index vector of model 1
ind= [1 2 3 4]';

% Transition matrix for the continous-time velocity model.
F = [0 0 1 0;
        0 0 0 1;
        0 0 0 0;
        0 0 0 0];

% Noise effect matrix for the continous-time system.
L = [0 0;
        0 0;
        1 0;
        0 1];

% Process noise variance
q1 = 1;
Qc= diag([q1 q1]);

% Discretization of the continous-time system.
[A,Q] = lti_disc(F,L,Qc,dt);

% Process noise variance for model 1 using EKF
EKF_q1 = 1;
EKF_Qc1 = diag([EKF_q1 EKF_q1]);

% Discretization of the continous-time system.
[EKF_A1,EKF_Q1] = lti_disc(F,L,EKF_Qc1,dt);

% Number of data points
n = 600;

% cpu time record
EKFtime = 0;
UKFtime = 0;

EKStime = 0;
UKStime = 0;

delayFtime = 0;
filterFtime = 0;
smoothFtime = 0;
predictFtime = 0;

MCruns = 2; T = 1:n; predictStep = 5;
SensorNoise_Factor = 1; %  using factor 1, just normal processing with everything perfect assumed by the estimator
% using the other to change the real sensor noise without informing any estimators

for rep = 1: MCruns
    % Space for real states and modes
    X_r = zeros(dims,n);

    %%%%%%% Creation of trajectory %%%%%%%
    % Start with constant velocity 1 toward right
    X_r(:,1) = [-12, 4, 15, -5]';

    % Generate object state values
    for i = 2:n
       st = 1;
       if isstr(a_func) | strcmp(class(a_func),'function_handle')
           X_r(ind,i) = feval(a_func,X_r(ind,i-1),a_param);
       else 
           X_r(ind,i) = A*X_r(ind,i-1);
       end
    end

    % Positions of sensors
    % $$$ S1 = [-0.5; 3];
    % $$$ S2 = [-0.5;-3];
    % $$$ S3 = [   7;-3];
    % $$$ S4 = [   7; 3];

    S1 = [-10; 0];
    S2 = [ 10; 0];

    s = [S1 S2 ]; % S3 S4

    % Handles to measurement models
    H = @bot_dh_dx;

    h_fun = @bot_h;

    h_param = s;


    % Standard deviation
    sd = 0.017*ones(1,size(s,2));
    R= diag(sd.^2);
    % Generate measurements
    Y = bot_h(X_r,s);
    % Add noise
    for i = 1:size(Y,2)
        st = 1;
        for j = 1:size(Y,1)
            Y(j,i) = Y(j,i) + sqrt(R(j,j)) * randn/SensorNoise_Factor; % !! Adapting the sensor noise without telling ESTIMATORS
        end
    end

%     % Print the trajectory
%     if print_figures
%         h = plot(X_r(1,:),X_r(2,:),'-g',...
%                  s(1,:),s(2,:),'k^',... 
%                  X_r(1,1),X_r(2,1),'ro','MarkerSize',12);    
%         legend('Real trajectory',...
%                'Positions of sensors',...
%                'Starting position', 'Location', 'North');
%         set(h,'markersize',5);
%         set(h,'linewidth',0.5);
%         set(gca,'FontSize',8);
%         xlim([-1 7.5]) 
%         ylim([-4 4]) 
%         if save_figures
%             print('-depsc','eimm2_trajectory.eps');
%         end
%         pause
%     end


    m = [-5 3 14 -6]';
    P = diag([1.5 0.05 0.3 0.01])^2;

    %% Space for the estimates.

    % EKF with model 1
    EKF_MM = zeros(size(A,1), size(Y,2));
    EKF_PP = zeros(size(A,1), size(A,1), size(Y,2));

    % UKF with model 1
    UKF_MM = zeros(size(A,1), size(Y,2));
    UKF_PP = zeros(size(A,1), size(A,1), size(Y,2));

    %%% Initial estimates %%%

    % EKF with model 1
    EKF_M = m;
    EKF_P = P;

    % UKF with model 1
    UKF_M = m;
    UKF_P = P;

    % Filtering steps.
    tic
    for i = T
        % EKF with model 1
        [EKF_M,EKF_P] = kf_predict(EKF_M,EKF_P,EKF_A1,EKF_Q1);
        [EKF_M,EKF_P] = ekf_update1(EKF_M,EKF_P,Y(:,i),H,R,h_fun,[],h_param);

        EKF_MM(:,i)   = EKF_M;
        EKF_PP(:,:,i) = EKF_P;
    end
    EKFtime = EKFtime+ toc;
    
    tic
    for i = T
        % UKF with model 1
        [UKF_M,UKF_P] = kf_predict(UKF_M,UKF_P,EKF_A1,EKF_Q1);
        [UKF_M,UKF_P] = ukf_update1(UKF_M,UKF_P,Y(:,i),h_fun,R,h_param);

        UKF_MM(:,i)   = UKF_M;
        UKF_PP(:,:,i) = UKF_P;
    end
    UKFtime = UKFtime+ toc;
    
%% smooth
    % Smooth the EKF estimates with RTS smoother using model 1
    tic
    [SM1,SP1] = rts_smooth(EKF_MM,EKF_PP,EKF_A1,EKF_Q1);
    EKStime = EKStime+toc;

    % Smooth the UKF estimates with RTS smoother using model 1
    tic
    [SM2,SP2] = rts_smooth(UKF_MM,UKF_PP,EKF_A1,EKF_Q1);
    UKStime = UKStime+toc;

    warning off
%% STF Fitting 
% parameter
    Parameter.initial = m; Parameter.Band = 10; Parameter.predictStep = predictStep;
% 2nd order fitting function; larger Parameter.Band will be required
% Fun2 = @(x,t) [atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,1),x(1)+x(2)*t+x(3)*t.^2-s(1,1));
%     atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,2),x(1)+x(2)*t+x(3)*t.^2-s(1,2));
%     atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,3),x(1)+x(2)*t+x(3)*t.^2-s(1,3)); 
%     atan2(x(4)+x(5)*t+x(6)*t.^2-s(2,4),x(1)+x(2)*t+x(3)*t.^2-s(1,4));]; %
% %     each row represents a bearing only sensor - 2-order ; 

% 1st order fitting function
Fun1 = @(x,t) [atan2(x(3)+x(4)*t-s(2,1),x(1)+x(2)*t-s(1,1));
    atan2(x(3)+x(4)*t-s(2,2),x(1)+x(2)*t-s(1,2));]; % each row represents a bearing only sensor - 1st order
    tic
    xfit_delay = nlinSTFfitting(Y,Fun1,'smoothing',Parameter); % 
    delayFtime =  toc;
    tic
    xfit_filter = nlinSTFfitting(Y,Fun1,'filtering',Parameter);
    filterFtime = toc;
    tic
    xfit_predict = nlinSTFfitting(Y,Fun1,'forecasting',Parameter);
    predictFtime = toc;
    tic
    Parameter.Band = 7; Parameter.FitOrder = 1; % reduce somehow the fitting iterval
    xfit_smooth = STFfitting(xfit_delay(:,end:-1:1),'smoothing',Parameter); % smoothing in time-reverse way
    xfit_smooth = xfit_smooth(:,end:-1:1);
    smoothFtime = smoothFtime + toc;

% Calculate the MSEs
    MSE_EKF(rep,:) =  (X_r(1,:)-EKF_MM(1,:)).^2+ (X_r(2,:)-EKF_MM(2,:)).^2;
    MSE_EKS(rep,:) = (X_r(1,:)-SM1(1,:)).^2 + (X_r(2,:)-SM1(2,:)).^2;
    MSE_UKF(rep,:) =  (X_r(1,:)-UKF_MM(1,:)).^2+ (X_r(2,:)-UKF_MM(2,:)).^2;
    MSE_UKS(rep,:) = (X_r(1,:)-SM2(1,:)).^2 + (X_r(2,:)-SM2(2,:)).^2;
    
    
    MSE_delayF(rep,:) =(X_r(1,:)-xfit_delay(1,:)).^2 + (X_r(2,:)-xfit_delay(2,:)).^2;
    MSE_filterF(rep,:) = (X_r(1,:)-xfit_filter(1,:)).^2+ (X_r(2,:)-xfit_filter(2,:)).^2;
    MSE_predictF(rep,:) = (X_r(1,Parameter.predictStep+1:end)-xfit_predict(1,Parameter.predictStep+1:end-Parameter.predictStep)).^2+(X_r(2,Parameter.predictStep+1:end)-xfit_predict(2,Parameter.predictStep+1:end-Parameter.predictStep)).^2 ;
    MSE_smoothF(rep,:) = (X_r(1,:)-xfit_smooth(1,:)).^2+ (X_r(2,:)-xfit_smooth(2,:)).^2;   
end

meanRMSE_EKF = sqrt(mean(MSE_EKF));
meanRMSE_EKS = sqrt(mean(MSE_EKS));
meanRMSE_UKF = sqrt(mean(MSE_UKF));
meanRMSE_UKS = sqrt(mean(MSE_UKS));


meanRMSE_delayF = sqrt(mean(MSE_delayF));
meanRMSE_filterF = sqrt(mean(MSE_filterF));
meanRMSE_predictF = sqrt(mean(MSE_predictF));
meanRMSE_smoothF = sqrt(mean(MSE_smoothF));


fprintf('============== The sensor noise is divided by SensorNoise_Factor= %.2f\n',SensorNoise_Factor);
fprintf('Only when SensorNoise_Factor == 1: all estimators are provided with the correct sensor statistics');


fprintf('Root Mean square errors of position estimates:\n');
fprintf('EKF-RMSE = %.4f\n',mean(meanRMSE_EKF));
fprintf('EKS-RMSE = %.4f\n',mean(meanRMSE_EKS));
fprintf('UKF-RMSE = %.4f\n',mean(meanRMSE_UKF));
fprintf('UKS-RMSE = %.4f\n',mean(meanRMSE_UKS));

fprintf('Online Fitting RMSE = %.4f\n',mean(meanRMSE_filterF));
fprintf('Delayed Fitting RMSE = %.4f\n',mean(meanRMSE_delayF));
fprintf('Smoothed Fitting RMSE = %.4f\n',mean(meanRMSE_smoothF));
fprintf('Fitting-based Forecasting RMSE = %.4f\n',mean(meanRMSE_predictF));


fprintf('Computing time:\n');
fprintf('EKF-time = %.4f\n',EKFtime/MCruns);
fprintf('EKS-time = %.4f\n',(EKFtime+EKStime)/MCruns);
fprintf('UKF-time = %.4f\n',UKFtime/MCruns);
fprintf('UKS-time = %.4f\n',(UKFtime+UKStime)/MCruns);

fprintf('Online Fitting time = %.4f\n',filterFtime/MCruns);
fprintf('Delayed Fitting time = %.4f\n',delayFtime/MCruns);
fprintf('Smoothed Fitting time = %.4f\n',(smoothFtime+delayFtime)/MCruns);
fprintf('Fitting-based Forecasting time = %.4f\n',predictFtime/MCruns);


% Plot the final filtering and smoothing results
figure
subplot(2,2,1),h0 = title('Truth'); hold on
h1 = plot(X_r(1,:),X_r(2,:),'k-',X_r(1,1),X_r(2,1),'r^',X_r(1,end),X_r(2,end),'rs',s(1,:),s(2,:),'o','markersize',4,'linewidth',2);
legend(h1,'True Trajectory','Trjajectory Starting Point','Trajectory Ending Point','Bearing-only Sensor'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
xlim([-1 7.5]); ylim([-3.5 3.5]) 

% figure
subplot(2,2,2),h0 = title('Filtering'); hold on
plot(X_r(1,:),X_r(2,:),'k-')
h1 = plot(EKF_MM(1,:),EKF_MM(2,:),'g--',UKF_MM(1,:),UKF_MM(2,:),'r--', xfit_filter(1,:),xfit_filter(2,:),'b-','markersize',2,'linewidth',2);
legend(h1,'EKF','UKF','Online Fitting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
xlim([-1 7.5]); ylim([-3.5 3.5]) 

% figure
subplot(2,2,3),h0 = title('Smoothing');hold on
plot(X_r(1,:),X_r(2,:),'k-'); 
h1 = plot(SM1(1,:),SM1(2,:),'--g',SM2(1,:),SM2(2,:),'--r',xfit_delay(1,:),xfit_delay(2,:),'r-',xfit_smooth(1,:),xfit_smooth(2,:),'b-','markersize',2,'linewidth',2);%
legend(h1,'EKF Smoothing','UKF Smoothing','Delayed Fitting','Smoothed Fitting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman'); %
xlim([-1 7.5]); ylim([-3.5 3.5]) 

% figure
subplot(2,2,4),h0 = title('Forecasting'); hold on
plot(X_r(1,:),X_r(2,:),'k-'),%
h1 = plot(xfit_predict(1,Parameter.predictStep+1:end),xfit_predict(2,Parameter.predictStep+1:end),'b->','markersize',2,'linewidth',2);
legend(h1,'Fitting-based 5-step Forecasting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
xlim([-1 7.5]); ylim([-3.5 3.5]) 

figure

h0=plot(T,meanRMSE_EKF,'g--',T,meanRMSE_UKF,'k--',T,meanRMSE_filterF,'b--',...
    T,meanRMSE_EKS,'g-',T,meanRMSE_UKS,'k-',T,meanRMSE_delayF,'r-',T,meanRMSE_smoothF,'b-',...
    T(predictStep+1:end),meanRMSE_predictF,'b:'); % 
set(gca,'FontSize',12);set(h0,'LineWidth',2);
h1=xlabel('Sampling steps'); h2=ylabel('RMSE');
h3=legend('EKF','UKF','Online Fitting','EKF Smoothing','UKF Smoothing','Delayed Fitting','Smoothed Fitting',...
  'Fitting-based 5-step Forecasting', 'location','best');%,'Orientation','horizontal'
set([h1,h2],'FontSize',14,'FontName','Times new Roman');%set(h3,'FontSize',12,'FontName','Times new Roman');
