%IMM_DEMO  Tracking a Target with Simple Manouvers demonstration
% 
% Simple demonstration for linear IMM using the following models:
%  1. Standard Wiener process velocity model
%  2. Standard Wiener process acceleration model
%
% The measurement model is linear, which gives noisy measurements 
% of target's position.
% 
% Copyright (C) 2007-2008 Jouni Hartikainen
%
% This software is distributed under the GNU General Public 
% Licence (version 2 or later); please refer to the file 
% Licence.txt, included with the software, for details.


%% 4 Fitting approaches are added by Tiancheng Li (t.c.li@usal.es)

% Continuous time trajectory Function of Time (FoT) fitting based 
%              Smoothing (Delayed fitting):  STFfittingDelay  - line 295
%              Filtering (Online fitting):   STFfittingOnline - line 298
%              Forecasting:                  STFfittingPredict -  line 301
%          and Smoothed fitting  :           line 305

% For the technical details, please refer to the paper:
% T. Li, H. Chen, S. Sun and J.M. Corchado, Joint Smoothing, Tracking, and Forecasting Based
% on Continuous-Time Target Trajectory Fitting, arXiv:1708.02196 [stat.AP]
% or the website: https://sites.google.com/site/tianchengli85/o2

%%
% Simple demonstration for IMM using velocity and acceleration models
clear
% Save figures
print_figures = 1;
save_figures = 1;


% Dimensionality of the state space
dims = 6;

nmodels = 2;

% Space for models
ind = cell(1,2);
A = cell(1,2);
Q = cell(1,2);
H = cell(1,2);
R = cell(1,2);

% Stepsize
dt = 0.1;

ind{2} = [1 2 3 4 5 6]';

% Transition matrix for the continous-time acceleration model.
F2 = [0 0 1 0 0 0;
      0 0 0 1 0 0;
      0 0 0 0 1 0;
      0 0 0 0 0 1;
      0 0 0 0 0 0;
      0 0 0 0 0 0];

% Noise effect matrix for the continous-time system.
L2 = [0 0;
      0 0;
      0 0;
      0 0;
      1 0;
      0 1];

% Process noise variance
q2 = 1.00;
Qc2 = diag([q2 q2]);

% Discretization of the continous-time system.
[A{2},Q{2}] = lti_disc(F2,L2,Qc2,dt);

ind{1} = [1 2 3 4]';

% Transition matrix for the continous-time velocity model.
F1 = [0 0 1 0;
      0 0 0 1;
      0 0 0 0;
      0 0 0 0];

% Noise effect matrix for the continous-time system.
L1 = [0 0;
      0 0;
      1 0;
      0 1];

% Process noise variance
q1 = .01;
Qc1 = diag([q1 q1]);

% Discretization of the continous-time system.
[A{1},Q{1}] = lti_disc(F1,L1,Qc1,dt);


H{1} = [1 0 0 0;
        0 1 0 0];

% Measurement models.
H{2} = [1 0 0 0 0 0;
        0 1 0 0 0 0];

hdims = 2;

% Variance in the measurements.
r1 = .1;
r2 = .1;
R{1} = diag([r1 r2]);

r1 = .1;
r2 = .1;
R{2} = diag([r1 r2]);

% cpu time record
KF1time = 0;
KF2time = 0;
IMMtime = 0;

KS1time = 0;
KS2time = 0;
IMMStime = 0;

IMMpretime = 0;

delayFtime = 0;
filterFtime = 0;
smoothFtime = 0;
predictFtime = 0;

MCruns = 100; predictStep = 5;
for rep = 1: MCruns
    
    % Generate the data.
    n = 200; 
    Y = zeros(hdims,n);
    X_r = zeros(dims,n);
    X_r(:,1) = [0 0 0 -1 0 0]';
    mstate = zeros(1,n);

    mu_ip = [0.9 0.1];
    mu_0j = mu_ip;
    %p_ij = [0.65 0.35;
    %        0.10 0.90];

    p_ij = [0.98 0.02;
            0.02 0.98];
%     p_ij = [0.9 0.1;
%             0.1 0.9];

    % $$$ % Generate random mode transitions
    % $$$ mstate(1) = 2;
    % $$$ for i = 2:n
    % $$$     r = rand;
    % $$$     for j = size(p_ij,1)-1;
    % $$$         if r < p_ij(mstate(i-1),j);
    % $$$             mstate(i) = j;
    % $$$         end
    % $$$     end
    % $$$     if mstate(i) == 0
    % $$$         mstate(i) = size(p_ij,1);
    % $$$     end
    % $$$ end

    % Forced mode transitions 
    mstate(1:50) = 1;
    mstate(51:70) = 2;
    mstate(71:120) = 1;
    mstate(121:150) = 2;
    mstate(151:200) = 1;


    % Acceleration model
    for i = 2:n
        st = mstate(i);
        X_r(ind{st},i) = A{st}*X_r(ind{st},i-1) + gauss_rnd(zeros(size(A{st},1),1), Q{st});
    end

    % Generate the measurements.
    for i = 1:n
        Y(:,i) = H{mstate(i)}*X_r(ind{mstate(i)},i) + gauss_rnd(zeros(size(Y,1),1), R{mstate(i)});
    end

    ac = floor(n/2)+1:floor(n/2)+2;
    clf; clc;
    %fprintf('Filtering with KF...');

    % $$$ plot(X_r(1,:),X_r(2,:),Y(1,:),Y(2,:),'.',X_r(1,1),...
    % $$$      X_r(2,1),'ro','MarkerSize',12);
    % $$$ legend('Real trajectory', 'Measurements');
    % $$$ title('Position');

    m = [0 0 0 -1 0 0]';
    P = diag([0.1 0.1 0.1 0.1 0.5 0.5]);

    %%%% Space for the estimates %%%% 

    % KF using model 1 
    KF1 = zeros(size(A{1},1), size(Y,2));
    PP1 = zeros(size(A{1},1), size(A{1},1), size(Y,2));

    % KF using model 2
    KF2 = zeros(size(A{2},1), size(Y,2));
    PP2 = zeros(size(A{2},1), size(A{2},1), size(Y,2));

    % Overall estimates of IMM filter
    MM = zeros(size(m,1), size(Y,2));
    PP = zeros(size(m,1), size(m,1), size(Y,2));
    % Model-conditioned estimates of IMM
    MM_i = cell(2,n);
    PP_i = cell(2,n);

    % Model probabilities 
    MU = zeros(2,size(Y,2));

    %%%% Prior estimates %%%%

    % KF1
    M1 = [0 0 0 -1]';
    P1 = diag([0.1 0.1 0.1 0.1]);

    % KF2
    M2 = [0 0 0 -1 0 0]';
    P2 = diag([0.1 0.1 0.1 0.1 0.5 0.5]);


    % IMM
    x_ip{1} = [0 0 0 -1]';
    P_ip{1} = diag([0.1 0.1 0.1 0.1]);
    x_ip{2} = [0 0 0 -1 0 0]';
    P_ip{2} = diag([0.1 0.1 0.1 0.1 0.5 0.5]);

    % Filtering steps.
    tic
    for i = 1:n
        % KF with model 1
        [M1,P1] = kf_predict(M1,P1,A{1},Q{1});
        [M1,P1] = kf_update(M1,P1,Y(:,i),H{1},R{1});
        KF1(:,i) = M1;
        PP1(:,:,i) = P1;
    end
    KF1time = KF1time + toc;
    
    tic
    for i = 1:n
        % KF with model 2
        [M2,P2] = kf_predict(M2,P2,A{2},Q{2});
        [M2,P2] = kf_update(M2,P2,Y(:,i),H{2},R{2});
        KF2(:,i) = M2;
        PP2(:,:,i) = P2;
    end
    KF2time = KF2time + toc;
    
UA = cell(3,n);  
    tic
    for i = 1:n
        % IMM
        [x_p,P_p,c_j] = imm_predict(x_ip,P_ip,mu_ip,p_ij,ind,dims,A,Q);
        [x_ip,P_ip,mu_ip,m,P] = imm_update(x_p,P_p,c_j,ind,dims,Y(:,i),H,R);
        MM(:,i)   = m;
        PP(:,:,i) = P;
        MU(:,i)   = mu_ip';
        MM_i(:,i) = x_ip';
        PP_i(:,i) = P_ip';
        UA{1,i}= x_ip; UA{2,i} = P_ip; UA{3,i}= mu_ip; 
    end
    IMMtime = IMMtime + toc;

MMpre = zeros(size(m,1), size(Y,2)+predictStep);    
    tic
    for i = 1:n
        a = UA{1,i}; b = UA{2,i}; c= UA{3,i}; 
        for tt = 1:predictStep
            [a,b,c,XX,PP] = imm_predict(a,b,c,p_ij,ind,dims,A,Q);
        end
        MMpre(:,predictStep+i) = XX;
    end
    IMMpretime = IMMpretime + toc;

    % Smooth the filtered estimates
    tic
    [SmKF1,SP1,SS1] = rts_smooth(KF1,PP1,A{1},Q{1});
    KS1time = KS1time + toc;
    tic
    [SmKF2,SP2,SS2] = rts_smooth(KF2,PP2,A{2},Q{2});
    KS2time = KS2time + toc;
    tic
    [SMM3,SP3,SM3_i,SP3_i,MU_S] = imm_smooth(MM,PP,MM_i,PP_i,MU,p_ij,mu_0j,ind,dims,A,Q,R,H,Y);
    IMMStime = IMMStime + toc;

    
    % STF Fitting parameter
    Parameter.Band = 10; Parameter.predictStep = predictStep; Parameter.FitOrder = 1;
    % the result is close to setting Parameter.Band = 20; Parameter.FitOrder = 2;
    
    % these following three fitting functions are jointly written as one in STFfitting; divided here for save computing
    tic
    xfit_delay = STFfittingDelay(Y,Parameter); % Equivalent to  xfit_delay = STFfitting(Y,'smoothing',Parameter)
    delayFtime = delayFtime + toc;
    tic
    xfit_filter = STFfittingOnline(Y,Parameter);% Equivalent to  xfit_filter = STFfitting(Y,'filtering',Parameter)
    filterFtime = filterFtime + toc;
    tic
    xfit_predict = STFfittingPredict(Y,Parameter);% Equivalent to  xfit_predict = STFfitting(Y,'forecasting',Parameter)
    predictFtime = predictFtime + toc;
    tic
    Parameter.Band = 7; % reduce somehow the fitting iterval
    xfit_smooth = STFfittingDelay(xfit_delay(:,end:-1:1),Parameter); % smoothing in time-reverse way; smoothed fitting
    xfit_smooth = xfit_smooth(:,end:-1:1);
    smoothFtime = smoothFtime + toc;

% Calculate the MSEs
    MSE_KF1(rep,:) = (X_r(1,:)-KF1(1,:)).^2 + (X_r(2,:)-KF1(2,:)).^2;
    MSE_KS1(rep,:) =  (X_r(1,:)-SmKF1(1,:)).^2 +(X_r(2,:)-SmKF1(2,:)).^2 ;
    MSE_KF2(rep,:) = (X_r(1,:)-KF2(1,:)).^2 + (X_r(2,:)-KF2(2,:)).^2;
    MSE_KS2(rep,:) =  (X_r(1,:)-SmKF2(1,:)).^2 + (X_r(2,:)-SmKF2(2,:)).^2;
    MSE_IMM(rep,:) =  (X_r(1,:)-MM(1,:)).^2 + (X_r(2,:)-MM(2,:)).^2;
    MSE_IMMS(rep,:) =  (X_r(1,:)-SMM3(1,:)).^2 + (X_r(2,:)-SMM3(2,:)).^2;
    
    MSE_IMMpre(rep,:) =  (X_r(1,predictStep+1:end)-MMpre(1,predictStep+1:end-predictStep)).^2 + (X_r(2,predictStep+1:end)-MMpre(2,predictStep+1:end-predictStep)).^2;
    % fitting
    MSE_delayF(rep,:)  =  (X_r(1,:)-xfit_delay(1,:)).^2 +(X_r(2,:)-xfit_delay(2,:)).^2;
    MSE_filterF(rep,:)  =  (X_r(1,:)-xfit_filter(1,:)).^2+(X_r(2,:)-xfit_filter(2,:)).^2;
    MSE_predictF(rep,:)  =  (X_r(1,predictStep+1:end)-xfit_predict(1,predictStep+1:end-predictStep)).^2 + (X_r(2,predictStep+1:end)-xfit_predict(2,predictStep+1:end-predictStep)).^2;
    MSE_smoothF(rep,:)  = (X_r(1,:)-xfit_smooth(1,:)).^2+ (X_r(2,:)-xfit_smooth(2,:)).^2;
    
%     fprintf('The number runs = %.4f\n',rep);
end


meanRMSE_KF1 = sqrt(mean(MSE_KF1));
meanRMSE_KF2 = sqrt(mean(MSE_KF2));
meanRMSE_KS1 = sqrt(mean(MSE_KS1));
meanRMSE_KS2 = sqrt(mean(MSE_KS2));
meanRMSE_IMM = sqrt(mean(MSE_IMM));
meanRMSE_IMMS = sqrt(mean(MSE_IMMS));
meanRMSE_IMMpre = sqrt(mean(MSE_IMMpre));

meanRMSE_delayF = sqrt(mean(MSE_delayF));
meanRMSE_filterF = sqrt(mean(MSE_filterF));
meanRMSE_smoothF = sqrt(mean(MSE_smoothF));
meanRMSE_predictF = sqrt(mean(MSE_predictF));

fprintf('Root Mean square errors of position estimates:\n');
fprintf('KF1-RMSE = %.4f\n',mean(meanRMSE_KF1));
fprintf('KS1-RMSE = %.4f\n',mean(meanRMSE_KS1));
fprintf('KF2-RMSE = %.4f\n',mean(meanRMSE_KF2));
fprintf('KS2-RMSE = %.4f\n',mean(meanRMSE_KS2));
fprintf('IMM-RMSE = %.4f\n',mean(meanRMSE_IMM));
fprintf('IMMS-RMSE = %.4f\n',mean(meanRMSE_IMMS));
fprintf('IMMpre-RMSE = %.4f\n',mean(meanRMSE_IMMpre));


fprintf('Online Fitting RMSE = %.4f\n',mean(meanRMSE_filterF));
fprintf('Delayed Fitting RMSE = %.4f\n',mean(meanRMSE_delayF));
fprintf('Smoothed Fitting RMSE = %.4f\n',mean(meanRMSE_smoothF));
fprintf('Fitting-based Forecasting RMSE = %.4f\n',mean(meanRMSE_predictF));


fprintf('Computing time:\n');
fprintf('KF1-time = %.4f\n',KF1time/MCruns);
fprintf('KS1-time = %.4f\n',(KF1time+KS1time)/MCruns);
fprintf('KF2-time = %.4f\n',KF2time/MCruns);
fprintf('KS2-time = %.4f\n',(KF2time+KS2time)/MCruns);
fprintf('IMM-time = %.4f\n',IMMtime/MCruns);
fprintf('IMMS-time = %.4f\n',(IMMtime+IMMStime)/MCruns);
fprintf('IMMpre-time = %.4f\n',(IMMtime+IMMpretime)/MCruns);

fprintf('Online Fitting time = %.4f\n',filterFtime/MCruns);
fprintf('Delayed Fitting time = %.4f\n',delayFtime/MCruns);
fprintf('Smoothed Fitting time = %.4f\n',(smoothFtime+delayFtime)/MCruns);
fprintf('Fitting-based Forecasting time = %.4f\n',predictFtime/MCruns);


save data_imm_demo
% Plot the final filtering and smoothing results

figure
subplot(2,2,1),h0=title('Truth');hold on
h1 = plot(X_r(1,:),X_r(2,:),'k-',Y(1,:),Y(2,:),'g.',X_r(1,1),X_r(2,1),'r^',X_r(1,end),X_r(2,end),'rs','markersize',6,'linewidth',2);
legend(h1,'True Trajectory','Observations','Trjajectory Starting Point','Trajectory Ending Point'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
% figure
subplot(2,2,2),h0=title('Filtering'); hold on
plot(X_r(1,:),X_r(2,:),'k-')
h1 = plot(KF1(1,:),KF1(2,:),'g--',KF2(1,:),KF2(2,:),'c--',MM(1,:),MM(2,:),'m--',xfit_filter(1,:),xfit_filter(2,:),'b--','markersize',2,'linewidth',2);
legend(h1,'KF (WPV)','KF (WPA)','IMM','Online Fitting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');
% figure
subplot(2,2,3),h0=title('Smoothing');hold on,set(h0,'FontSize',12,'FontName','Times new Roman');
plot(X_r(1,:),X_r(2,:),'k-');
h1 = plot(SmKF1(1,:),SmKF1(2,:),'-g',SmKF2(1,:),SmKF2(2,:),'-c',SMM3(1,:),SMM3(2,:),'-m',xfit_delay(1,:),xfit_delay(2,:),'r-',xfit_smooth(1,:),xfit_smooth(2,:),'b-','markersize',2,'linewidth',2); %
legend(h1,'KF (WPV) Smoothing','KF (WPA) Smoothing','IMM Smoothing','Delayed Fitting','Smoothed Fitting');set([gca, h0],'FontSize',12,'FontName','Times new Roman');
% figure
subplot(2,2,4),h0=title('Forecasting');hold on,
plot(X_r(1,:),X_r(2,:),'k-'), 
h1 = plot(MMpre(1,predictStep+1:end),MMpre(2,predictStep+1:end),'r:',xfit_predict(1,Parameter.predictStep+1:end),xfit_predict(2,Parameter.predictStep+1:end),'b:','markersize',2,'linewidth',2);
legend(h1,'IMM-based 5-step Forecasting','Fitting-based 5-step Forecasting'); set([gca, h0],'FontSize',12,'FontName','Times new Roman');

figure
T = 1:n; 
h0=plot(T,meanRMSE_KF1,'g--',T,meanRMSE_KF2,'c--',T,meanRMSE_IMM,'m--',T,meanRMSE_filterF,'b--',...
    T,meanRMSE_KS1,'g-',T,meanRMSE_KS2,'c-',T,meanRMSE_IMMS,'m-',T,meanRMSE_delayF,'r-',T,meanRMSE_smoothF,'b-',...
    T(predictStep+1:end),meanRMSE_IMMpre,'r:',T(predictStep+1:end),meanRMSE_predictF,'b:');%,
set(gca,'FontSize',12);set(h0,'LineWidth',2);
h1=xlabel('Sampling steps'); h2=ylabel('RMSE');
h3=legend('KF (WPV)','KF (WPA)','IMM','Online Fitting','KF (WPV) Smoothing','KF (WPA) Smoothing','IMM Smoothing','Delayed Fitting','Smoothed Fitting',...
   'IMM-based 5-step Forecasting','Fitting-based 5-step Forecasting', 'location','best');%,'Orientation','horizontal'
set([h1,h2],'FontSize',14,'FontName','Times new Roman');%set(h3,'FontSize',12,'FontName','Times new Roman');


% if print_figures
%     
% %     figure
% %     h = plot(Y(1,:),Y(2,:),'g*',...
% %              X_r(1,:),X_r(2,:),'k-',...         
% %              KF1(1,:),KF1(2,:),'c-',...
% %              SmKF1(1,:),SmKF1(2,:),'m-',...
% %              xfit_smooth(1,:),xfit_smooth(2,:),'r.-',...
% %              xfit_filter(1,:),xfit_filter(2,:),'b.-');
% %     legend('Measurement',...
% %            'True trajectory',...
% %            'Filtered',...
% %            'Smoothed',...
% %            'History-Smoothed fitting O2',...
% %            'Online fitting O2 output');
% %     title('Estimates produced by Kalman filter using the model 1.');
% %     set(h,'markersize',2);
% %     set(h,'linewidth',2);
% %     set(gca,'FontSize',12);
% %     if save_figures
% %         print('-depsc','imm1.eps');
% %     end
% %     figure
% %     h = plot(Y(1,:),Y(2,:),'g*',...
% %              X_r(1,:),X_r(2,:),'k-',...         
% %              KF2(1,:),KF2(2,:),'c-',...
% %              SmKF2(1,:),SmKF2(2,:),'m-',...
% %              xfit_smooth(1,:),xfit_smooth(2,:),'r.-',...
% %              xfit_filter(1,:),xfit_filter(2,:),'b.-');
% %     legend('Measurement',...
% %            'True trajectory',...
% %            'Filtered',...
% %            'Smoothed',...
% %            'History-Smoothed fitting O2',...
% %            'Online fitting O2 output');
% %     title('Estimates produced by Kalman filter using the model 2.');
% %     set(h,'markersize',2);
% %     set(h,'linewidth',2);
% %     set(gca,'FontSize',12);
% %     if save_figures
% %         print('-depsc','imm2.eps');
% %     end
%     figure
%     h = plot(Y(1,:),Y(2,:),'g.',...
%              X_r(1,:),X_r(2,:),'k-',...         
%              MM(1,:),MM(2,:),'y-',...
%              KF2(1,:),KF2(2,:),'c-',...
%              SmKF2(1,:),SmKF2(2,:),'m-',...
%              xfit_smooth(1,:),xfit_smooth(2,:),'r.-',...
%              xfit_filter(1,:),xfit_filter(2,:),'b.-');
%     legend('Measurement',...
%            'True trajectory',...
%            'IMM',...
%            'KF model 2',...
%            'Smoothing KF model 2',...
%            'Batch fitting O2',...
%            'Online fitting O2');
%     title('Estimates produced by IMM-filter.')
%     set(h,'markersize',2);
%     set(h,'linewidth',2);
%     set(gca,'FontSize',12);
%     if save_figures
%         print('-depsc','imm3.eps');
%     end
%     
% %     h = plot(1:n,2-mstate,'g--',...
% %              1:n,MU(1,:)','r-',...  
% %              1:n,MU_S(1,:)','b-');         
% %     legend('True',...
% %            'Filtered',...
% %            'Smoothed');
% %     title('Probability of model 1');
% %     ylim([-0.1,1.1]);
% %     set(h,'markersize',2);
% %     set(h,'linewidth',0.5);
% %     set(gca,'FontSize',6);
% %     if save_figures
% %         print('-depsc','imm4.eps');
% %     end
% end
